# ruff: noqa: F401


from .api import (
    AnyJobStepConfig,
    GeospatialPartitionJobStepConfig,
    JobConfig,
    JobStepConfig,
    NonGeospatialPartitionJobStepConfig,
    PartitionJobStepConfig,
    RootAnyJobStepConfig,
    UdfAccessToken,
    UdfAccessTokenList,
    UdfJobStepConfig,
)
from .internal import JobResponse, Jobs, RunResponse
from .schema import (
    DataType,
    Field,
    FixedSizeBinaryType,
    FixedSizeListType,
    LargeListType,
    ListType,
    PrimitiveDataType,
    Schema,
    StructType,
)
from .udf import GeoPandasUdfV2, Header



# ruff: noqa: F401

from . import api, models, types, warnings
from ._load_udf import load
from ._options import env as _env
from ._options import options
from ._public_api import ingest, ingest_nongeospatial
from ._run import run
from ._secrets import secrets
from ._submit import submit
from ._udf import udf
from ._utils import utils

# TODO: fused.{download, get_chunk_from_table, get_chunks_metadata} are deprecated here
from .core import cache, download, file_path, get_chunk_from_table, get_chunks_metadata
from .ipython.fused_magics import load_ipython_extension

# Version is set to 0.0.0 as the version string is automatically generated by
# https://github.com/mtkennerly/poetry-dynamic-versioning
__version__ = "1.20.0"

__all__ = [
    "api",
    "cache",
    "download",
    "file_path",
    "get_chunk_from_table",
    "get_chunks_metadata",
    "ingest",
    # "ingest_nongeospatial",
    "load",
    # "models",
    "run",
    # "secrets",
    "submit",
    # "types",
    "udf",
    "utils",
    # "warnings",
    # "options",
    # "_env",
    "load_ipython_extension",
]



from pathlib import Path
from typing import Any, Union

import requests

from fused.models.api.udf_access_token import is_udf_token
from fused.models.udf import AnyBaseUdf

from .core import (
    load_udf_from_code,
    load_udf_from_directory,
    load_udf_from_file,
    load_udf_from_fused,
    load_udf_from_github,
    load_udf_from_shared_token,
    load_udf_from_zip,
)


def load(
    url_or_udf: Union[str, Path],
    /,
    *,
    cache_key: Any = None,
    import_globals: bool = True,
) -> AnyBaseUdf:
    """
    Loads a UDF from various sources including GitHub URLs,
    and a Fused platform-specific identifier.

    This function supports loading UDFs from a GitHub repository URL, or a Fused
    platform-specific identifier composed of an email and UDF name. It intelligently
    determines the source type based on the format of the input and retrieves the UDF
    accordingly.

    Args:
        url_or_udf: A string representing the location of the UDF, or the raw code of the UDF.
            The location can be a GitHub URL starting with "https://github.com",
            a Fused platform-specific identifier in the format "email/udf_name",
            or a local file path pointing to a Python file.
        cache_key: An optional key used for caching the loaded UDF. If provided, the function
            will attempt to load the UDF from cache using this key before attempting to
            load it from the specified source. Defaults to None, indicating no caching.
        import_globals: Expose the globals defined in the UDF's context as attributes on the UDF object (default True).
            This requires executing the code of the UDF. To globally configure this behavior, use `fused.options.never_import`.

    Returns:
        AnyBaseUdf: An instance of the loaded UDF.

    Raises:
        ValueError: If the URL or Fused platform-specific identifier format is incorrect or
            cannot be parsed.
        Exception: For errors related to network issues, file access permissions, or other
            unforeseen errors during the loading process.

    Examples:
        Load a UDF from a GitHub URL:
        ```py
        udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/REM_with_HyRiver/")
        ```

        Load a UDF using a Fused platform-specific identifier:
        ```py
        udf = fused.load("username@fused.io/REM_with_HyRiver")
        ```
    """
    # NOTE: Order matters
    if str(url_or_udf).startswith("https://github.com"):
        return load_udf_from_github(
            url_or_udf, cache_key=cache_key, import_globals=import_globals
        )

    if isinstance(url_or_udf, str) and "\n" in url_or_udf:
        # if there are newlines, assume this is raw code
        return load_udf_from_code(url_or_udf, import_globals=import_globals)

    path = Path(url_or_udf)

    if path.suffix == ".py":
        return load_udf_from_file(path, import_globals=import_globals)

    if path.suffix == ".zip":
        return load_udf_from_zip(path, import_globals=import_globals)

    if path.is_dir():  # Non-existant paths will be false
        return load_udf_from_directory(path, import_globals=import_globals)

    if isinstance(url_or_udf, str):
        if is_udf_token(url_or_udf):
            try:
                return load_udf_from_shared_token(
                    url_or_udf, import_globals=import_globals
                )
            except requests.HTTPError:
                raise ValueError(
                    "It looks like you tried to load a UDF from a shared token "
                    "without having access to the source code. "
                    "You may want to call `run` instead of `load`."
                )

        # remaining cases: assume UDF name or user/name
        parts = url_or_udf.split("/", maxsplit=1)
        if len(parts) == 2:
            email_or_handle, udf_name = parts
            return load_udf_from_fused(
                email_or_handle,
                udf_name,
                cache_key=cache_key,
                import_globals=import_globals,
            )
        elif len(parts) == 1:
            udf_name = parts[0]
            return load_udf_from_fused(
                udf_name, cache_key=cache_key, import_globals=import_globals
            )

    if not path.exists():
        raise FileNotFoundError(str(path))

    else:
        raise ValueError(
            f"Ambiguous file detected {path}. Ensure file exists and has an extension."
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any

import requests
from pydantic import BaseModel, ConfigDict

from fused._formatter.formatter_eval_result import fused_eval_result_repr
from fused.models.schema import Schema
from fused.models.udf import AnyBaseUdf


class CacheSources(str, Enum):
    MEMORY = "memory"
    OBJECT_STORAGE = "object_storage"


@dataclass
class CacheStatus:
    is_cached: bool
    cache_source: CacheSources


def is_response_cached(r: requests.Response) -> CacheStatus:
    # Metadata is always persisted from original call and its cache=True value indicates that the UDF result was saved
    # in object storage, but doesn't tell us if the invocation was cached or regenerated and saved. Therefore, we read
    # the headers set by entities that know, either fused server or the cache proxy.
    from_cached_storage = r.headers.get("x-fused-cached") is not None
    from_cached_proxy = r.headers.get("x-cache", "").lower() == "hit"
    is_cached = from_cached_storage or from_cached_proxy

    if from_cached_proxy:
        cache_source = CacheSources.MEMORY
    elif from_cached_storage:
        cache_source = CacheSources.OBJECT_STORAGE
    else:
        cache_source = None

    return CacheStatus(is_cached=is_cached, cache_source=cache_source)


class UdfEvaluationResult(BaseModel):
    data: Any = None
    sidecar: bytes | None = None

    udf: AnyBaseUdf | None = None
    table_schema: Schema | None = None

    time_taken_seconds: float

    stdout: str | None = None
    stderr: str | None = None
    has_exception: bool = False
    exception_class: str | None = None
    error_message: str | None = None
    error_lineno: int | None = None

    # TODO: Maybe update to one field cache_status: Optional[CacheStatus] = None ?
    is_cached: bool = False
    cache_source: CacheSources | None = None

    def _repr_html_(self) -> str:
        return fused_eval_result_repr(self)

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MultiUdfEvaluationResult(BaseModel):
    udf_results: list[UdfEvaluationResult | Any]

    def _repr_html_(self) -> str:
        # Aggregate reprs
        result_reprs = [
            udf_result._repr_html_()
            if hasattr(udf_result, "_repr_html_")
            else repr(udf_result)
            for udf_result in self.udf_results
        ]
        return "<br><br>".join(result_reprs)

    model_config = ConfigDict(arbitrary_types_allowed=True)



import inspect
import json
import uuid
from io import StringIO
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Union, get_args, get_origin

try:
    from types import UnionType
except ImportError:
    # compatibility with Python 3.9
    UnionType = type(Union[int, str])

from fused._optional_deps import (
    GPD_GEODATAFRAME,
    HAS_GEOPANDAS,
    HAS_MERCANTILE,
    HAS_PANDAS,
    HAS_SHAPELY,
    PD_DATAFRAME,
    SHAPELY_GEOMETRY,
)

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import geopandas as gpd
except ImportError:
    gpd = None

try:
    import shapely
except ImportError:
    shapely = None

try:
    import mercantile
except ImportError:
    mercantile = None


if TYPE_CHECKING:
    import geopandas as gpd
    import mercantile
    import pandas as pd
    import shapely

    from fused.types import Bbox, Bounds, TileGDF, TileXYZ


FALSE_CASEFOLDED = "false".casefold()
GEOJSON_CRS = 4326


def _parse_bbox(val: str) -> Union[list, "gpd.GeoDataFrame"]:
    deserialized = json.loads(val)
    if isinstance(deserialized, list):
        return deserialized
    if HAS_GEOPANDAS:
        return gpd.GeoDataFrame.from_features(deserialized, crs=GEOJSON_CRS)
    return val


def _parse_to_gdf(val: Any) -> "gpd.GeoDataFrame":
    if isinstance(val, gpd.GeoDataFrame):
        return val
    elif isinstance(val, str):
        deserialized = json.loads(val)
        if isinstance(deserialized, list):
            return gpd.GeoDataFrame(
                {}, geometry=[shapely.box(*deserialized)], crs=GEOJSON_CRS
            )
        return gpd.GeoDataFrame.from_features(deserialized, crs=GEOJSON_CRS)
    raise NotImplementedError(f"Not sure how to convert type `{val}` to GeoDataFrame")


def _parse_to_df(val: Any) -> "pd.DataFrame":
    if isinstance(val, pd.DataFrame):
        return val
    elif isinstance(val, str):
        return pd.read_json(StringIO(val))
    raise NotImplementedError(f"Not sure how to convert type `{val}` to DataFrame")


def convert_to_tile_gdf(val: Any, add_xyz: bool = True) -> "TileGDF":
    if not HAS_GEOPANDAS:
        raise ImportError(
            "GeoPandas is not installed, but is required for running a UDF with "
            "a TileGDF or ViewportGDF parameter."
        )
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, list):
        val = gpd.GeoDataFrame({}, geometry=[shapely.box(*val)], crs=GEOJSON_CRS)
    if isinstance(val, gpd.GeoDataFrame):
        if (
            add_xyz
            and not all(c in val.columns for c in ["x", "y", "z"])
            and HAS_MERCANTILE
        ):
            tile = mercantile.bounding_tile(*val.total_bounds)
            val = val.assign(x=tile.x, y=tile.y, z=tile.z)
            val = val[["x", "y", "z", "geometry"]]
        return val

    raise NotImplementedError(f"Not sure how to convert type `{val}` to TileGDF")


def _parse_to_geometry(val: Any) -> "shapely.Geometry":
    if not HAS_SHAPELY:
        raise ImportError(
            "shapely is not installed, but is required for running a UDF with "
            "a Geometry parameter."
        )
    if isinstance(val, str):
        return shapely.from_wkt(val)
    if isinstance(val, shapely.Geometry):
        return val
    raise NotImplementedError(
        f"Not sure how to convert type `{val}` to shapely.Geometry"
    )


def convert_to_tile_xyz(val: Any) -> "TileXYZ":
    if not HAS_MERCANTILE:
        raise ImportError(
            "mercantile is not installed, but is required for running a UDF with "
            "a TileXYZ parameter."
        )
    if isinstance(val, str):
        val = _parse_bbox(val)
    if HAS_GEOPANDAS and isinstance(val, gpd.GeoDataFrame):
        if all(c in val.columns for c in ["x", "y", "z"]):
            return mercantile.Tile(*val.iloc[0][["x", "y", "z"]])
        else:
            return mercantile.bounding_tile(*val.total_bounds)
    if isinstance(val, list):
        return mercantile.bounding_tile(*val)
    raise NotImplementedError(f"not sure how to convert type `{val}` to TileXYZ")


def convert_to_bounds(val: Any) -> "Bounds":
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, list):
        if not len(val) == 4:
            raise ValueError(f"Expected 4 values for bounds, got {len(val)}")
        return val
    if isinstance(val, gpd.GeoDataFrame):
        return val.total_bounds.tolist()
    raise NotImplementedError(f"not sure how to convert type `{val}` to Bounds")


def convert_to_bbox(val: Any) -> "Bbox":
    if not HAS_SHAPELY:
        raise ImportError(
            "shapely is not installed, but is required for running a UDF with "
            "a Bbox parameter."
        )
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, list):
        return shapely.box(*val)
    if isinstance(val, gpd.GeoDataFrame):
        return shapely.box(*val.total_bounds)
    raise NotImplementedError(f"not sure how to convert type `{val}` to Bbox")


def _resolve_annotation(param: inspect.Parameter | None):
    annotation = param.annotation if param else inspect._empty
    if annotation is inspect._empty:
        return annotation

    origin = get_origin(annotation)
    if origin is not None:
        if origin is Union or origin is UnionType:
            args = get_args(annotation)
            # Handle Optional and (x | None)
            if len(args) == 2:
                if args[0] is type(None):
                    return args[1]
                if args[1] is type(None):
                    return args[0]
            # TODO, handle e.g. Union types. Look into using Pydantic for this. Leave as is for now.
        else:
            return origin
    return annotation


def coerce_arg(
    val: Any, param: inspect.Parameter | None, default_annotation: Any | None = None
) -> Any:
    """
    Args:
        val: The value to coerce
        param: How the parameter is defined on the function, or None if there is no corresponding
               parameter (which happens when passing kwargs).
        default_annotation: If the parameter annotation is empty or the parmeter is None, a default
                            annotation to use as a backup. E.g. for kwargs allows setting them as
                            `str` without affecting the parsing of named parameters with empty
                            annotations.
    """
    # Needed due to circular import issue
    from fused.types import Bbox, Bounds, Tile, TileGDF, TileXYZ, ViewportGDF

    # TODO: perhaps use https://docs.python.org/3/library/typing.html#typing.get_type_hints
    annotation = _resolve_annotation(param)
    if annotation is inspect._empty:
        if default_annotation:
            # Default unannotated parameters and kwargs to str
            annotation = default_annotation
        else:
            return val

    # Temporary solution for properly handling non-stdlib annotations
    if isinstance(annotation, str):
        if "GeoDataFrame" in annotation and HAS_GEOPANDAS:
            annotation = GPD_GEODATAFRAME
        elif "DataFrame" in annotation and HAS_PANDAS:
            annotation = PD_DATAFRAME

    if annotation is str:
        return str(val)
    if annotation is int:
        if isinstance(val, str):
            # base 0 will cause the integer to be interpreted as a integer literal similar to how
            # code would be read. Possible improvement to help handle 0x...:
            # https://docs.python.org/3/library/functions.html#int
            # Only strings should be passed in to `int` if the base is passed too. Integers and so
            # on will not be converted correctly.
            return int(val, base=10)
        return int(val)
    if annotation is float:
        return float(val)
    if annotation is bool:
        if isinstance(val, str) and val.casefold() == FALSE_CASEFOLDED:
            return False
        return bool(val)
    if annotation in [list, dict, List, Dict, Iterable] and isinstance(val, str):
        return json.loads(val)
    if annotation is tuple:
        return tuple(json.loads(val))
    if annotation is uuid.UUID:
        return uuid.UUID(val)
    if annotation in [Tile, TileGDF, ViewportGDF]:
        return convert_to_tile_gdf(val, add_xyz=annotation in [Tile, TileGDF])
    if annotation is TileXYZ:
        return convert_to_tile_xyz(val)
    if annotation is Bounds:
        return convert_to_bounds(val)
    if annotation is Bbox:
        return convert_to_bbox(val)
    if HAS_GEOPANDAS and annotation is GPD_GEODATAFRAME:
        # load as geojson
        return _parse_to_gdf(val)
    if HAS_PANDAS and annotation is PD_DATAFRAME:
        # load as json
        return _parse_to_df(val)
    if (
        HAS_SHAPELY
        and isinstance(annotation, type)
        and issubclass(annotation, SHAPELY_GEOMETRY)
    ):
        return _parse_to_geometry(val)

    # Unknown, fall back to not doing anything with it
    return val



from __future__ import annotations

import io
import os
import warnings
from datetime import timedelta
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Literal,
    overload,
)

from fused._global_api import get_api
from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._options import options as OPTIONS
from fused._secrets import secrets
from fused.models.api import ListDetails
from fused.models.api._cronjob import CronJob
from fused.models.internal import Jobs
from fused.models.internal.job import CoerceableToJobId, RunResponse
from fused.models.udf import BaseUdf
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.warnings import FusedTypeWarning

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def job_get_logs(
    job: CoerceableToJobId,
    since_ms: int | None = None,
) -> list[Any]:
    """Fetch logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

    Returns:
        Log messages for the given job.
    """
    api = get_api()
    return api.get_logs(job=job, since_ms=since_ms)


def job_print_logs(
    job: CoerceableToJobId,
    since_ms: int | None = None,
    file: IO | None = None,
) -> None:
    """Fetch and print logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.
        file: Where to print logs to. Defaults to sys.stdout.

    Returns:
        None
    """
    job = RunResponse.from_job_id(job)
    job.print_logs(since_ms=since_ms, file=file)


def job_tail_logs(
    job: CoerceableToJobId,
    refresh_seconds: float = 1,
    sample_logs: bool = True,
    timeout: float | None = None,
):
    """Continuously print logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
        sample_logs: if true, print out only a sample of logs. Defaults to True.
        timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
    """
    api = get_api()
    return api.tail_logs(
        job=job,
        refresh_seconds=refresh_seconds,
        sample_logs=sample_logs,
        timeout=timeout,
    )


def job_get_status(job: CoerceableToJobId) -> RunResponse:
    """Fetch the status of a running job

    Args:
        job: the identifier of a job or a `RunResponse` object.

    Returns:
        The status of the given job.
    """
    api = get_api()
    return api.get_status(job=job)


def job_cancel(job: CoerceableToJobId) -> RunResponse:
    """Cancel an existing job

    Args:
        job: the identifier of a job or a `RunResponse` object.

    Returns:
        A new job object.
    """
    api = get_api()
    return api.cancel_job(job=job)


def job_get_exec_time(job: CoerceableToJobId) -> timedelta:
    """Determine the execution time of this job, using the logs.

    Returns:
        Time the job took. If the job is in progress, time from first to last log message is returned.
    """
    job = RunResponse.from_job_id(job)
    return job.get_exec_time()


def job_wait_for_job(
    job: CoerceableToJobId,
    poll_interval_seconds: float = 5,
    timeout: float | None = None,
) -> RunResponse:
    """Block the Python kernel until this job has finished

    Args:
        poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
        timeout: The length of time in seconds to wait for the job. Defaults to None.

    Raises:
        TimeoutError: if waiting for the job timed out.

    Returns:
        The status of the given job.
    """
    api = get_api()
    return api.wait_for_job(
        job,
        poll_interval_seconds=poll_interval_seconds,
        timeout=timeout,
    )


def job_wait_for_results(
    job: CoerceableToJobId,
    poll_interval_seconds: float = 5,
    timeout: float | None = None,
) -> list[UdfEvaluationResult]:
    job_response = job_get_status(job=job)
    return job_response.wait_for_job(
        poll_interval_seconds=poll_interval_seconds, timeout=timeout
    ).get_results()


def job_get_results(job: CoerceableToJobId) -> list[Any]:
    return job_get_status(job=job).get_results()


def whoami():
    """
    Returns information on the currently logged in user
    """
    api = get_api()
    return api._whoami()


def delete(
    path: str,
    max_deletion_depth: int | Literal["unlimited"] = 3,
) -> bool:
    """Delete the files at the path.

    Args:
        path: Directory or file to delete, like `fd://my-old-table/`
        max_deletion_depth: If set (defaults to 3), the maximum depth the operation will recurse to.
                            This option is to help avoid accidentally deleting more data that intended.
                            Pass `"unlimited"` for unlimited.


    Examples:
        ```python
        fused.api.delete("fd://bucket-name/deprecated_table/")
        ```
    """
    api = get_api()
    return api.delete(path, max_deletion_depth=max_deletion_depth)


@overload
def list(path: str, *, details: Literal[False] = False) -> list[str]: ...


@overload
def list(path: str, *, details: Literal[True]) -> list[ListDetails]: ...


def list(path: str, *, details: bool = False) -> list[str] | list[ListDetails]:
    """List the files at the path.

    Args:
        path: Parent directory URL, like `fd://bucket-name/`
        details: If True, return additional metadata about each record.

    Returns:
        A list of paths as URLs, or as metadata objects.

    Examples:
        ```python
        fused.api.list("fd://bucket-name/")
        ```
    """
    api = get_api(credentials_needed=False)
    return api.list(path, details=details)


def get(path: str) -> bytes:
    """Download the contents at the path to memory.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`

    Returns:
        bytes of the file

    Examples:
        ```python
        fused.api.get("fd://bucket-name/file.parquet")
        ```
    """
    api = get_api(credentials_needed=False)
    return api.get(path)


def download(path: str, local_path: str | Path) -> None:
    """Download the contents at the path to disk.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`
        local_path: Path to a local file.
    """
    api = get_api(credentials_needed=False)
    api.download(path, local_path=local_path)


def upload(
    local_path: str | Path | bytes | BinaryIO | pd.DataFrame | gpd.GeoDataFrame,
    remote_path: str,
    timeout: float | None = None,
) -> None:
    """Upload local file to S3.

    Args:
        local_path: Either a path to a local file (`str`, `Path`), a (Geo)DataFrame
                    (which will get uploaded as Parquet file), or the contents to upload.
                    Any string will be treated as a Path, if you wish to upload the contents of
                    the string, first encode it: `s.encode("utf-8")`
        remote_path: URL to upload to, like `fd://new-file.txt`
        timeout: Optional timeout in seconds for the upload (will default to `OPTIONS.request_timeout` if not specified).

    Examples:
        To upload a local json file to your Fused-managed S3 bucket:
        ```py
        fused.api.upload("my_file.json", "fd://my_bucket/my_file.json")
        ```
    """
    api = get_api()
    if isinstance(local_path, str):
        # We assume any string being passed in is a path, rather than the contents
        # to upload.
        local_path = Path(local_path)
        if not local_path.exists():
            warnings.warn(
                FusedTypeWarning(
                    '`local_path` is being treated as a path but it does not exist. If you wish to upload the contents of the string, encode it to bytes first with `.encode("utf-8")'
                )
            )
    if isinstance(local_path, Path):
        file_name = local_path.name
        if remote_path.endswith("/"):  # Remote path has a directory-like structure
            remote_path += file_name

        data = local_path.read_bytes()
    elif HAS_PANDAS and isinstance(local_path, PD_DATAFRAME):
        data = io.BytesIO()
        local_path.to_parquet(data)
        data.seek(0)
    else:
        data = local_path

    api.upload(path=remote_path, data=data, timeout=timeout)


def sign_url(path: str) -> str:
    """Create a signed URL to access the path.

    This function may not check that the file represented by the path exists.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`

    Returns:
        HTTPS URL to access the file using signed access.

    Examples:
        ```python
        fused.api.sign_url("fd://bucket-name/table_directory/file.parquet")
        ```
    """
    api = get_api()
    return api.sign_url(path)


def sign_url_prefix(path: str) -> dict[str, str]:
    """Create signed URLs to access all blobs under the path.

    Args:
        path: URL to a prefix, like `fd://bucket-name/some_directory/`

    Returns:
        Dictionary mapping from blob store key to signed HTTPS URL.

    Examples:
        ```python
        fused.api.sign_url_prefix("fd://bucket-name/table_directory/")
        ```
    """
    api = get_api()
    return api.sign_url_prefix(path)


def get_jobs(
    n: int = 5,
    *,
    skip: int = 0,
    per_request: int = 25,
    max_requests: int | None = None,
) -> Jobs:
    """Get the job history.

    Args:
        n: The number of jobs to fetch. Defaults to 5.

    Keyword Args:
        skip: Where in the job history to begin. Defaults to 0, which retrieves the most recent job.
        per_request: Number of jobs per request to fetch. Defaults to 25.
        max_requests: Maximum number of requests to make. May be None to fetch all jobs. Defaults to 1.

    Returns:
        The job history.
    """
    api = get_api()
    return api.get_jobs(
        n=n, skip=skip, per_request=per_request, max_requests=max_requests
    )


def get_udfs(
    n: int = 10,
    *,
    skip: int = 0,
    per_request: int = 25,
    max_requests: int | None = None,
    by: Literal["name", "id", "slug"] = "name",
    whose: Literal["self", "public"] = "self",
) -> dict:
    """
    Fetches a list of UDFs.

    Args:
        n: The total number of UDFs to fetch. Defaults to 10.
        skip: The number of UDFs to skip before starting to collect the result set. Defaults to 0.
        per_request: The number of UDFs to fetch in each API request. Defaults to 25.
        max_requests: The maximum number of API requests to make.
        by: The attribute by which to sort the UDFs. Can be "name", "id", or "slug". Defaults to "name".
        whose: Specifies whose UDFs to fetch. Can be "self" for the user's own UDFs or "public" for
            UDFs available publicly. Defaults to "self".

    Returns:
        A list of UDFs.

    Examples:
        Fetch UDFs under the user account:
        ```py
        fused.api.get_udfs()
        ```
    """
    api = get_api()
    return api.get_udfs(
        n=n,
        skip=skip,
        per_request=per_request,
        max_requests=max_requests,
        by=by,
        whose=whose,
    )


def get_apps(
    n: int = 10,
    *,
    skip: int = 0,
    per_request: int = 25,
    max_requests: int | None = None,
    by: Literal["name", "id", "slug"] = "name",
    whose: Literal["self", "public"] = "self",
) -> dict:
    """
    Fetches a list of apps.

    Args:
        n: The total number of apps to fetch. Defaults to 10.
        skip: The number of apps to skip before starting to collect the result set. Defaults to 0.
        per_request: The number of apps to fetch in each API request. Defaults to 25.
        max_requests: The maximum number of API requests to make.
        by: The attribute by which to sort the apps. Can be "name", "id", or "slug". Defaults to "name".
        whose: Specifies whose apps to fetch. Can be "self" for the user's own apps or "public" for
            UDFs available publicly. Defaults to "self".

    Returns:
        A list of apps.

    Examples:
        Fetch apps under the user account:
        ```py
        fused.api.get_apps()
        ```
    """
    api = get_api()
    return api.get_apps(
        n=n,
        skip=skip,
        per_request=per_request,
        max_requests=max_requests,
        by=by,
        whose=whose,
    )


def enable_gcs():
    """
    Save gcs credentials from AWS secret manager into a temporary local file and set its path to the env variable.
    """
    gcs_secret = secrets[OPTIONS.gcs_secret]
    if gcs_secret:
        with open(OPTIONS.gcs_filename, "w") as f:
            f.write(gcs_secret)

        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = OPTIONS.gcs_filename


def schedule_udf(
    udf: BaseUdf | str,
    minute: list[int] | int,
    hour: list[int] | int,
    day_of_month: list[int] | int | None = None,
    month: list[int] | int | None = None,
    day_of_week: list[int] | int | None = None,
    udf_args: dict[str, Any] | None = None,
    enabled: bool = True,
    _create_udf: bool = True,
    **kwargs,
) -> CronJob:
    """Schedule a UDF to run on a cron schedule.

    Args:
        udf: The UDF to schedule.
        minute: The minute to run the UDF on.
        hour: The hour to run the UDF on.
        day_of_month: The day of the month to run the UDF on. (Default every day)
        month: The month to run the UDF on. (Default every month)
        day_of_week: The day of the week to run the UDF on. (Default every day)
        udf_args: The arguments to pass to the UDF. (Default None)
        enabled: Whether the cron job is enabled. (Default True)
        _create_udf: Save the UDF to Fused before creating the CronJob. (Default True)
    """
    return CronJob.from_udf(
        udf,
        minute,
        hour,
        day_of_month,
        month,
        day_of_week,
        udf_args,
        enabled,
        _create_udf=_create_udf,
        **kwargs,
    )


def schedule_list():
    """List all cron jobs"""
    return CronJob.list_team()



from datetime import datetime
from typing import Any, Iterator, Sequence

from pydantic import Field, RootModel, field_validator

from fused._global_api import get_api
from fused.models.base import FusedBaseModel
from fused.models.internal import RunResponse
from fused.models.udf import BaseUdf
from fused.models.udf.base_udf import METADATA_FUSED_DESCRIPTION, METADATA_FUSED_ID


class CronJob(FusedBaseModel):
    """A scheduled run of the UDF"""

    id: str | None = None
    """ID of the cron job"""
    name: str
    """Name of the cron job"""
    description: str | None = None
    execution_environment_id: str | None = None
    """ID of the team that owns the cron job"""
    enabled: bool
    minute: list[int]
    hour: list[int]
    day_of_month: list[int]
    month: list[int]
    day_of_week: list[int]
    udf_id: str = Field(alias="udf")
    """ID of the UDF to run"""
    udf_args: dict[str, Any] | None = None
    """Arguments to pass to the UDF"""
    last_run: datetime | None = None
    """The last time the cron job was run"""

    @field_validator("minute", mode="before")
    @classmethod
    def validate_minute(cls, v):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, list) or not all(0 <= i <= 59 for i in v):
            raise ValueError("Minute must be between 0 and 59")
        if len(v) > 1:
            raise ValueError("Minute must be a single integer")
        return v

    @field_validator("hour", mode="before")
    @classmethod
    def validate_hour(cls, v):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, list) or not all(0 <= i < 24 for i in v):
            raise ValueError("Hour must be between 0 and 23")
        return v

    @field_validator("day_of_month", mode="before")
    @classmethod
    def validate_day_of_month(cls, v):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, list) or not all(1 <= i <= 31 for i in v):
            raise ValueError("Day of month must be between 1 and 31")
        return v

    @field_validator("month", mode="before")
    @classmethod
    def validate_month(cls, v):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, list) or not all(1 <= i <= 12 for i in v):
            raise ValueError("Month must be between 1 and 12")
        return v

    @field_validator("day_of_week", mode="before")
    @classmethod
    def validate_day_of_week(cls, v):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, list) or not all(0 <= i <= 6 for i in v):
            raise ValueError("Day of week must be between 0 and 6")
        return v

    @classmethod
    def from_udf(
        cls,
        udf: BaseUdf | str,
        minute: list[int] | int,
        hour: list[int] | int,
        day_of_month: list[int] | int | None = None,
        month: list[int] | int | None = None,
        day_of_week: list[int] | int | None = None,
        udf_args: dict[str, Any] | None = None,
        enabled: bool = True,
        description: str | None = None,
        _create_udf: bool = True,
        **kwargs,
    ) -> "CronJob":
        """Create a cron job from a UDF"""
        from fused._load_udf import load as fused_load

        if not isinstance(udf, BaseUdf):
            udf = fused_load(udf)

        udf_id = udf._get_metadata_safe(METADATA_FUSED_ID)
        if _create_udf and udf_id is None:
            try:
                udf = udf.to_fused()
                udf_id = udf._get_metadata_safe(METADATA_FUSED_ID)
            except Exception as e:
                raise ValueError(
                    "Could not save UDF. Call to_fused() on the UDF."
                ) from e
        elif udf_id is None:
            raise ValueError(
                "UDF has not been saved yet, call to_fused() on the UDF or CronJob.from_udf with _create_udf=True"
            )

        day_of_month = day_of_month if day_of_month is not None else list(range(1, 32))
        month = month if month is not None else list(range(1, 13))
        day_of_week = day_of_week if day_of_week is not None else list(range(7))

        cronjob = cls(
            udf_id=udf_id,
            name=udf.name,
            description=description
            or udf._get_metadata_safe(METADATA_FUSED_DESCRIPTION),
            enabled=enabled,
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month=month,
            day_of_week=day_of_week,
            udf_args=udf_args,
            **kwargs,
        )
        return cronjob.create()

    @classmethod
    def from_id(cls, cronjob_id: str) -> "CronJob":
        """Get a cron job by it's ID"""
        return get_api().get_cronjob(cronjob_id=cronjob_id)

    @classmethod
    def list_team(cls) -> list["CronJob"]:
        """List all cron jobs for the team"""
        return get_api().list_cronjobs()

    @property
    def udf(self) -> BaseUdf:
        """Get the UDF that this cron job will run"""
        from fused._load_udf import load as fused_load

        return fused_load(self.udf_id)

    @property
    def _schedule(self) -> str:
        """Retrieve the schedule as a cron-compatible string."""

        def format_cron_part(
            part: list[int] | int, default: list[int] | None = None
        ) -> str:
            if isinstance(part, int):
                return str(part)
            elif len(part) == 1:
                return part[0]
            elif default is not None and part == default:
                return "*"
            else:
                return ",".join(str(i) for i in part)

        minute_part = format_cron_part(self.minute, list(range(60)))
        hour_part = format_cron_part(self.hour, list(range(24)))
        dom_part = format_cron_part(self.day_of_month, list(range(1, 32)))
        month_part = format_cron_part(self.month, list(range(1, 13)))
        dow_part = format_cron_part(self.day_of_week, list(range(7)))
        return f"{minute_part} {hour_part} {dom_part} {month_part} {dow_part}"

    def delete(self) -> "CronJob":
        """Delete the cron job, unscheduling it."""
        if self.id is None:
            raise ValueError("This cron job has not been saved yet")
        deleted_cronjob = self._api.delete_cronjob(self.id)
        self.id = None
        deleted_cronjob.id = None
        return deleted_cronjob

    def create(self) -> "CronJob":
        """Create the cron job, scheduling it to run."""
        new_cronjob = self._api.create_cronjob(self)
        self.id = new_cronjob.id
        self.execution_environment_id = new_cronjob.execution_environment_id
        return new_cronjob

    def update(self, create: bool = False) -> "CronJob":
        """Update the cron job, changing its schedule or arguments using the values in this object.

        Changes on this object will not take effect until this method is called.

        Args:
            create: If True, create the cron job if it has not been saved yet.
        """
        if self.id is None:
            if create:
                return self.create()
            else:
                raise ValueError(
                    "This cron job has not been saved yet. Call update(create=True) to save it."
                )

        return self._api.update_cronjob(self)

    def refresh_from_api(self) -> "CronJob":
        """Refresh the cron job from the saved version on the server"""
        if self.id is None:
            raise ValueError("This cron job has not been saved yet")
        return self._api.get_cronjob(self.id)

    def run_now(self) -> RunResponse:
        """Trigger a run of the cron job now"""
        run = self._api.run_cronjob(self.id)
        self.last_run = run["cronjob"].last_run
        return run["run"]

    def __repr__(self) -> str:
        disabled = " [disabled]" if not self.enabled else ""
        not_saved = " [not saved -- call create()]" if not self.id else ""
        return f"<CronJob {self.name} ({self._schedule}){disabled}{not_saved}>"


class CronJobSequence(RootModel[Sequence[CronJob]]):
    def _repr_html_(self) -> str:
        rows = "\n".join(
            f"""
            <tr>
                <td>{cronjob.name}</td>
                <td>{cronjob._schedule}{" (disabled)" if not cronjob.enabled else ""}{" (not saved)" if not cronjob.id else ""}</td>
                <td>{cronjob.last_run}</td>
            </tr>
        """
            for cronjob in self.root
        )
        return f"""
        <table>
            <tr>
                <th>Name</th>
                <th>Schedule</th>
                <th>Last run</th>
            </tr>
            {rows}
        </table>
        """

    def __iter__(self) -> Iterator[CronJob]:
        return iter(self.root)

    def __getitem__(self, item: int) -> CronJob:
        return self.root[item]

    def __setitem__(self, item: int, value: CronJob) -> None:
        self.root[item] = value

    def __len__(self) -> int:
        return len(self.root)

    def __repr__(self) -> str:
        return repr(self.root)



import inspect
import linecache
import warnings
from contextlib import ExitStack
from typing import Any, Callable, Iterable, Mapping

from fused.core._impl._context_impl import context_in_batch
from fused.models.udf import GeoPandasUdfV2, MockUdfInput, Output


def _compile_and_get_local(udf: GeoPandasUdfV2) -> Callable:
    """
    Compile and exec the UDF's code, returning a callable that can be used
    to run the UDF.

    - Compile the UDF code ensuring proper line numbers
    - Exec the code in a context that includes the `fused` module
    - Extract the entrypoint function from the exec'd code
    - Wrap the entrypoint function to include preprocessing of (keyword) arguments
      (coerce based on the type annotations in the UDF signature)
    """
    import fused
    from fused._udf.args import coerce_arg
    from fused.warnings import FusedDeprecationWarning

    code = udf.code
    local_name = udf.entrypoint

    # Ensure stack traces, getsource, etc work correctly with the exec'd code:
    compile_name = f"<udf {local_name}>"
    linecache.cache[compile_name] = (
        len(code),
        None,
        [line + "\n" for line in code.splitlines()],
        compile_name,
    )

    # Contextual reading: eval vs exec vs compile.
    # https://stackoverflow.com/questions/2220699/whats-the-difference-between-eval-exec-and-compile
    # mode="exec" is used to compile the code in the form of a module.
    compiled_python = compile(code, compile_name, mode="exec")

    # At this point, the compiled_python code could look like this in bytecode terms:
    #
    #   @fused.udf
    #   def udf():
    #       print("hello world")
    #
    # And when we exec the bytecode object, it injects the udf into the
    # exec_globals_locals below.
    def entrypoint_wrapper(*args, **kwargs):
        exec_globals_locals = {"fused": fused}
        """
        exec_globals_locals must be the same exact reference, or else the following will crash:
        In [10]: exec('''
        ...: import pandas as pd
        ...:
        ...: def aaa():
        ...:   print(pd.DataFrame())
        ...:
        ...: aaa()
        ...: ''', {}, {})
        See here for why exec_globals_locals must be the same reference:
        https://docs.python.org/3/library/functions.html#exec
        """
        exec(compiled_python, exec_globals_locals, exec_globals_locals)
        if local_name not in exec_globals_locals:
            raise NameError(
                f"Could not find {local_name}. You need to define a UDF with `def {local_name}()`."
            )
        # Reference to a callable or UDF object
        entrypoint = exec_globals_locals[local_name]

        # Ensure the entrypoint is something callable (not a UDF object)
        if isinstance(entrypoint, GeoPandasUdfV2):
            entrypoint = entrypoint._nested_callable

        kwargs_to_pass = udf.parameters or {}

        # Fixup parameters with special meaning
        # kwargs comes from try_running_partition_function in udf.py, this just confirms that since this function
        # doesn't have the parameters explicitly defined.
        signature = inspect.signature(entrypoint)
        param_names = signature.parameters.keys()
        if kwargs is not None:
            for key, value in kwargs.items():
                if key in param_names:
                    kwargs_to_pass[key] = value

        input_kwargs = {}
        input_as_udf_args = kwargs["input"].as_udf_args() if "input" in kwargs else {}
        for key, data in input_as_udf_args.items():
            param = signature.parameters.get(key)
            if param is not None:
                input_kwargs[key] = coerce_arg(data, param)

        # Temp fix for backwards compatibility: try to pass bbox as first positional param if 'bbox' arg not defined
        if (
            "bounds" not in input_kwargs
            and len(args) == 0
            and "bounds" in input_as_udf_args
            and len(signature.parameters) > 0
        ):
            bounds = input_as_udf_args["bounds"]
            param = None

            # first check for a bbox keyword argument
            if "bbox" in signature.parameters and "bbox" not in kwargs_to_pass:
                param = signature.parameters["bbox"]
                kwargs_to_pass["bbox"] = bounds

            # otherwise pass to first positional argument
            else:
                param = list(signature.parameters.values())[0]
                # Check to ensure that the user didn't already provide this value as a kwarg
                if param in kwargs_to_pass:
                    raise TypeError(f"Multiple values provided for argument '{param}'.")
                kwargs_to_pass[param.name] = coerce_arg(bounds, param)

            if param:
                warnings.warn(
                    "The tile bounds are being passed in through the argument "
                    f"'{param.name}'. This behavior is deprecated in favor of "
                    "using the 'bounds' argument explicitly. "
                    f"Consider renaming '{param.name}' to 'bounds'.",
                    FusedDeprecationWarning,
                )

        if context_in_batch():
            converted_args = convert_args_to_kwargs(entrypoint, args)
            kwargs_to_pass.update(converted_args)
        elif args:
            raise ValueError(
                "Unhandled UDF use-case passing args. Calling code needs to adapt to pass args as kwargs."
            )

        has_kwargs = any(
            [
                signature.parameters[param_name].kind == inspect.Parameter.VAR_KEYWORD
                for param_name in signature.parameters
            ]
        )
        reformatted_kwargs_to_pass = {
            key: coerce_arg(
                arg,
                signature.parameters.get(key),
                default_annotation=(
                    None if key in signature.parameters and key != "input" else str
                ),
            )
            for key, arg in kwargs_to_pass.items()
            # If this corresponds to a parameter on the UDF, then pass it in.
            # Otherwise, simply ignore it.
            if (key in signature.parameters and key != "input") or has_kwargs
        }

        return entrypoint(**input_kwargs, **reformatted_kwargs_to_pass)

    return entrypoint_wrapper


def convert_args_to_kwargs(udf_fn: Callable, args: tuple) -> dict:
    """Convert batch args into kwargs

    A single positional arg can only be passed when using arg_list. This function handles converting positional
    arguments to keyword arguments for the case where a user wants a single dict to be used as keyword arguments for
    the entire UDF. If the input dict does not match the UDF signature, we fall back to passing the dict to the first
    parameter.
    """
    signature = inspect.signature(udf_fn)
    kwargs = {}
    if args:
        # Treat dict values as kwargs if keys are a subset of the parameter names
        if isinstance(args[0], dict) and args[0].keys() <= signature.parameters.keys():
            return args[0]

        first_param_name = None
        for name in list(signature.parameters.keys()):
            first_param_name = name
            break
        if first_param_name:
            kwargs = {first_param_name: args[0]}

    return kwargs


def compile_udf_and_run_v2_simple(
    code: GeoPandasUdfV2,
    args: Iterable[Any] = (),
    kwargs: Mapping[str, Any] | None = None,
):
    from fused._udf.state import (
        decorator_src_override_context,
        decorator_udf_override_context,
    )

    with ExitStack() as stack:
        if code.headers is not None:
            for header in code.headers:
                stack.enter_context(header._register_custom_finder())

        # Without this the UDF object will only get the function it is wrapping.
        # Because then that fn cannot reference anythng else in the UDF.
        stack.enter_context(decorator_src_override_context(code.code))
        # Without this the UDF object will only get the code, but not any of the
        # other fields on the UDF object
        stack.enter_context(decorator_udf_override_context(code))

        entrypoint = _compile_and_get_local(code)
        kwargs_to_pass = kwargs if kwargs else {}
        return entrypoint(*args, **kwargs_to_pass)


def compile_udf_and_run_v3(code: GeoPandasUdfV2, input: MockUdfInput, output: Output):
    from fused._udf.state import (
        decorator_src_override_context,
        decorator_udf_override_context,
    )

    with ExitStack() as stack:
        # UDF's that are created inside a running UDF.
        if code.headers is not None:
            # Python modules.
            for header in code.headers:
                # Needed to hack imports.
                #
                # We inject the whole UDF object below so we do not need to
                # enter these headers into the context here.
                stack.enter_context(header._register_custom_finder())

        # Without this the UDF object will only get the function it is wrapping.
        # Because then that fn cannot reference anythng else in the UDF.
        stack.enter_context(decorator_src_override_context(code.code))
        # Without this the UDF object will only get the code, but not any of the
        # other fields on the UDF object
        stack.enter_context(decorator_udf_override_context(code))

        # entrypoint is either an object (wrapping another UDF) or a function.
        # Ensure that from this point onwards, entrypoint is always a function.
        entrypoint = _compile_and_get_local(code)
        result = entrypoint(input=input)

    # Support returning a DataFrame directly from the UDF.
    if result is not None:
        output.data = result
    return result


def compile_udf_and_run_with_kwargs(udf: GeoPandasUdfV2, *args, **kwargs):
    udf = udf.model_copy(deep=True)
    udf.parameters = {
        **(udf.parameters if udf.parameters else {}),
        **kwargs,
    }

    output = Output()
    if len(args) == 0:
        input = MockUdfInput(None)
    elif len(args) == 1:
        input = MockUdfInput(args[0])
    else:
        raise ValueError(
            "Unexpected number of positional arguments to UDF. Pass keyword arguments instead."
        )
    compile_udf_and_run_v3(udf, input=input, output=output)

    return output.data



class FusedWarning(Warning):
    pass


class FusedDeprecationWarning(FusedWarning):
    pass


class FusedInProgressWarning(FusedWarning):
    pass


class FusedNonProductionWarning(FusedWarning):
    pass


class FusedOnPremWarning(FusedWarning):
    pass


class FusedIPythonWarning(FusedWarning):
    pass


class FusedUdfWarning(FusedWarning):
    pass


class FusedDefaultWarning(FusedWarning):
    pass


class FusedPythonVersionWarning(FusedWarning):
    pass


class FusedTypeWarning(FusedWarning):
    pass


class FusedPathWarning(FusedWarning):
    pass


class FusedReservedWarning(FusedWarning):
    pass


class FusedIgnoredWarning(FusedWarning):
    pass


class FusedImportWarning(FusedWarning):
    pass


class FusedRefreshingWarning(FusedWarning):
    pass



from typing import Any, Dict


class MockUdfInput:
    def __init__(self, data: Any) -> None:
        self._data = data

    def as_udf_args(self) -> Dict[str, Any]:
        kwargs = {}
        if self._data is not None:
            kwargs["bounds"] = self._data

        return kwargs



"""Models to describe objects for input/output of a UDF"""

# ruff: noqa: F401

from .base_udf import BaseUdf, UdfType
from .header import Header
from .input import MockUdfInput
from .output import Output
from .udf import (
    EMPTY_UDF,
    AnyBaseUdf,
    GeoPandasUdfV2,
    RootAnyBaseUdf,
    load_udf_from_response_data,
)

__all__ = [
    "BaseUdf",
    "Header",
    "EMPTY_UDF",
    "AnyBaseUdf",
    "GeoPandasUdfV2",
    "RootAnyBaseUdf",
    "UdfType",
    "load_udf_from_response_data",
    "Output",
]



from typing import TYPE_CHECKING, Any, Union

from pydantic import BaseModel, ConfigDict

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd
    import pyarrow as pa

    DataType = Union[pa.Table, pd.DataFrame, gpd.GeoDataFrame, None]
else:
    DataType = Any


class Output(BaseModel):
    data: DataType = None

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")



from __future__ import annotations

import json
import warnings
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING
from zipfile import ZipFile

from fused._environment import infer_display_method
from fused._options import options as OPTIONS
from fused.warnings import FusedImportWarning, FusedWarning

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def deserialize_tiff(content: bytes):
    import rasterio
    import rioxarray
    import xarray as xr

    with NamedTemporaryFile(
        dir=OPTIONS.temp_directory, prefix="udf_result", suffix=".tiff"
    ) as ntmp:
        with open(ntmp.name, "wb") as f:
            f.write(content)
        with rasterio.open(ntmp.name) as dataset:
            rda = rioxarray.open_rasterio(
                dataset,
                masked=True,
            )
            # ensure we get the data back in memory (and not a reference to the temp file)
            # TODO rasterio should probably be able to directly read from in-memory bytes
            # instead of writing to a temp file
            rda = rda.load()
            metadata = dataset.tags(ns="fused")
        orig_type = metadata.get("orig_type")
        if orig_type == "numpy.ndarray":
            data = rda.values
            if len(json.loads(metadata["shape"])) < rda.ndim:
                data = data.squeeze()
            if "bounds" in metadata:
                bounds = json.loads(metadata["bounds"])
                data = (data, bounds)
        elif orig_type == "xarray.DataArray":
            data = rda
        else:
            key = metadata.get("key") or "image"
            data = xr.Dataset({key: rda})
    return data


def deserialize_png(content: bytes):
    import xarray as xr
    from PIL import Image

    display_method = infer_display_method(None, None)
    if display_method.show_widget:
        from IPython.display import Image as IPythonImage
        from IPython.display import display

        display(IPythonImage(data=content, format="png"))

    image = Image.open(BytesIO(content))
    width, height = image.size
    if len(image.getbands()) == 1:
        image_data = list(image.getdata())
        image_data = [
            image_data[i : i + width] for i in range(0, len(image_data), width)
        ]
        data_array = xr.DataArray(image_data, dims=["y", "x"])
    else:
        image_data = []
        for band in range(len(image.getbands())):
            band_data = list(image.getdata(band=band))
            band_data = [
                band_data[i : i + width] for i in range(0, len(band_data), width)
            ]
            image_data.append(band_data)
        data_array = xr.DataArray(image_data, dims=["band", "y", "x"])

    # Create the dataset with image, latitude, and longitude data
    dataset = xr.Dataset({"image": data_array})

    return dataset


def deserialize_parquet(content: bytes) -> "pd.DataFrame" | "gpd.GeoDataFrame":
    import pyarrow.parquet as pq

    output_df: pd.DataFrame | None = None

    meta = pq.read_metadata(BytesIO(content))
    if b"geo" in meta.metadata:
        try:
            import geopandas as gpd

            output_df = gpd.read_parquet(BytesIO(content))
        except ImportError:
            warnings.warn(
                "`geopandas` package is not installed so geometries are displayed as "
                "bytes instead of parsed shapes",
                FusedImportWarning,
            )
        except (ValueError, KeyError) as e:
            warnings.warn(
                f"Result has geo metadata but could not be loaded in GeoPandas: {e}",
                FusedWarning,
            )
    if output_df is None:
        try:
            import pandas as pd
        except ImportError:
            raise ModuleNotFoundError(
                "`pandas` package is not installed. Please install `pandas` to run "
                "this UDF which returns tabular data."
            )

        output_df = pd.read_parquet(BytesIO(content))

    # Force all column names to be strings
    output_df.columns = [str(x) for x in output_df.columns]

    return output_df


def deserialize_json(content: bytes):
    if content == b"":
        return None

    # TODO: Read X-Fused-Orig-Type header here too
    return json.loads(content)


def deserialize_zip(content: bytes):
    buf = BytesIO(content)
    deserialized = {}
    with ZipFile(buf, "r") as zf:
        meta_obj = json.loads(zf.read("meta.json"))

        for file, file_meta in meta_obj.get("files", {}).items():
            file_content = zf.read(file)
            deserialized[file] = parse_realtime_response_bytes(
                file_content, file_meta["content_type"]
            )

    return deserialized


def parse_realtime_response_bytes(content: bytes, content_type: str):
    if content_type == "application/octet-stream":  # parquet
        data = deserialize_parquet(content)
    elif content_type == "image/png":
        data = deserialize_png(content)
    elif content_type == "image/tiff":
        data = deserialize_tiff(content)
    elif content_type == "application/json":
        data = deserialize_json(content)
    elif content_type == "application/zip":
        data = deserialize_zip(content)
    else:
        data = content  # TODO

    return data



try:
    from IPython.core.magic import Magics, cell_magic, magics_class
except ImportError as e:
    import_error = e

    def load_ipython_extension(_):
        raise ImportError(f"IPython was not able to be imported: {import_error!r}")

else:
    # TODO: Update fused.load for raw source?
    from fused.core._udf_ops import load_udf_from_code

    @magics_class
    class FusedMagics(Magics):
        """Fused IPython Magics Collection"""

        # Function name is used as the magic name
        @cell_magic
        def udf(self, udf_name, udf_source):
            """Create a UDF out of the content of the cell block

            '%%udf udf_name' creates a UDF with the name 'udf_name'
            """
            udf = load_udf_from_code(udf_source, udf_name)
            self.shell.user_ns[udf_name] = udf

    def load_ipython_extension(ipython):
        ipython.register_magics(FusedMagics)



from __future__ import annotations

import ast
import warnings
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Iterable,
    Literal,
    Sequence,
    overload,
)

from pydantic import Field, PrivateAttr, RootModel
from typing_extensions import Annotated, Self, override

from fused.models.schema import Schema
from fused.models.udf.base_udf import (
    METADATA_FUSED_ID,
    METADATA_FUSED_SLUG,
    BaseUdf,
    UdfType,
)
from fused.warnings import FusedUdfWarning

from .._inplace import _maybe_inplace

if TYPE_CHECKING:
    from fused.models.api.job import UdfJobStepConfig
    from fused.models.udf._eval_result import UdfEvaluationResult


class GeoPandasUdfV2(BaseUdf):
    """A user-defined function that operates on [`geopandas.GeoDataFrame`s][geopandas.GeoDataFrame]."""

    type: Literal[UdfType.GEOPANDAS_V2] = UdfType.GEOPANDAS_V2
    table_schema: Schema | None = None
    """The [`Schema`][fused.models.Schema] describing the output of this UDF.
    """

    entrypoint: str
    """Name of the function within the code to invoke."""

    cache_max_age: int | None = None
    """The maximum age when returning a result from the cache."""

    parameters: dict[str, Any] = Field(default_factory=dict)
    """Parameters to pass into the entrypoint."""

    _parameter_list: Sequence[str] | None = PrivateAttr(None)
    _parameter_has_kwargs: bool | None = PrivateAttr(None)
    original_headers: str | None = None

    _nested_callable = PrivateAttr(None)  # TODO : Find out type

    def _with_udf_entrypoint(self) -> Self:
        """
        If the entrypoint is not "udf", replace it back to "udf" using AST manipulation.
        If a function named "udf" already exists, it will be renamed to "udf_updated".
        """
        target_new_name = "udf"
        if self.entrypoint == target_new_name:
            return self

        udf = self.model_copy()

        try:
            # Parse the code into an AST
            tree = ast.parse(udf.code, type_comments=True)

            # Rename existing udf -> udf_old
            visitor1 = RenameFunctionVisitor(
                old_name=target_new_name,
                new_name=f"{target_new_name}_old",
            )
            updated_tree = visitor1.visit(tree)

            # Rename the entrypoint -> udf
            visitor2 = RenameFunctionVisitor(
                old_name=self.entrypoint,
                new_name=target_new_name,
            )
            updated_tree = visitor2.visit(updated_tree)

            # Check if the original entrypoint function definition was found in the second pass
            if not visitor2.found_definition:
                raise ValueError(
                    f"Could not find the specified entrypoint function definition '{self.entrypoint}' in the UDF code."
                )

            # Ensure line numbers and column offsets are correct
            ast.fix_missing_locations(updated_tree)

            # Convert the potentially modified AST back to code
            udf.code = ast.unparse(updated_tree)
            udf.entrypoint = target_new_name

        except Exception as e:
            # Catch parsing, visiting, or unparsing errors
            raise RuntimeError(f"Failed to update entrypoint of UDF: {e}") from e

        return udf

    def set_parameters(
        self,
        parameters: dict[str, Any],
        replace_parameters: bool = False,
        inplace: bool = False,
    ) -> GeoPandasUdfV2:
        """Set the parameters on this UDF.

        Args:
            parameters: The new parameters dictionary.
            replace_parameters: If True, unset any parameters not in the parameters argument. Defaults to False.
            inplace: If True, modify this object. If False, return a new object. Defaults to True.
        """
        ret = _maybe_inplace(self, inplace)
        new_parameters = (
            parameters
            if replace_parameters
            else {
                **ret.parameters,
                **parameters,
            }
        )
        ret.parameters = new_parameters
        return ret

    def eval_schema(self, inplace: bool = False) -> GeoPandasUdfV2:
        """Reload the schema saved in the code of the UDF.

        Note that this will evaluate the UDF function.

        Args:
            inplace: If True, update this UDF object. Otherwise return a new UDF object (default).
        """
        from fused._udf.execute_v2 import execute_for_decorator

        new_udf = execute_for_decorator(self)
        assert isinstance(new_udf, GeoPandasUdfV2), (
            f"UDF has unexpected type: {type(new_udf)}"
        )
        ret = _maybe_inplace(self, inplace)
        ret.table_schema = new_udf.table_schema
        ret._parameter_list = new_udf._parameter_list
        ret._parameter_has_kwargs = new_udf._parameter_has_kwargs
        return ret

    def run_local(
        self,
        sample: Any | None = ...,
        *,
        inplace: bool = False,
        validate_imports: bool | None = None,
        **kwargs,
    ) -> UdfEvaluationResult:
        """Evaluate this UDF against a sample.

        Args:
            sample: Sample (from `get_sample`) to execute against.
            inplace: If True, update this UDF object with schema information. (default)
        """
        from fused._udf.execute_v2 import execute_against_sample

        ret = _maybe_inplace(self, inplace)
        sample_list = [] if sample is Ellipsis else [sample]
        return execute_against_sample(
            udf=ret,
            input=sample_list,
            validate_imports=validate_imports,
            **kwargs,
        )

    def to_file(self, where: str | Path | BinaryIO, *, overwrite: bool = False):
        """Write the UDF to disk or the specified file-like object.

        The UDF will be written as a Zip file.

        Args:
            where: A path to a file or a file-like object.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        updated_udf = self._with_udf_entrypoint()
        job = updated_udf()
        job.export(where, how="zip", overwrite=overwrite)

    def to_directory(self, where: str | Path | None = None, *, overwrite: bool = False):
        """Write the UDF to disk as a directory (folder).

        Args:
            where: A path to a directory. If not provided, uses the UDF function name.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        updated_udf = self._with_udf_entrypoint()
        job = updated_udf()
        where = where or self.name
        job.export(where, how="local", overwrite=overwrite)

    # List of data input is passed - run that
    @overload
    def __call__(self, *, arg_list: Iterable[Any], **kwargs) -> UdfJobStepConfig: ...

    # Nothing is passed - run the UDF once
    @overload
    def __call__(self, *, arg_list: None = None, **kwargs) -> UdfJobStepConfig: ...

    def __call__(
        self, *, arg_list: Iterable[Any] | None = None, **kwargs
    ) -> UdfJobStepConfig:
        """Create a job from this UDF.

        Args:
            arg_list: A list of records to pass in to the UDF as input.
        """
        # cyclic dependency if imported at top-level
        from fused.models.api.job import UdfJobStepConfig

        with_params = self.model_copy()
        # TODO: Consider using with_parameters here, and validating that "context" and other reserved parameter names are not being passed.
        new_parameters = {**kwargs}
        if new_parameters:
            with_params.parameters = new_parameters

        if arg_list is not None and not len(arg_list):
            warnings.warn(
                FusedUdfWarning(
                    "An empty `arg_list` was passed in, no calls to the UDF will be made."
                )
            )

        return UdfJobStepConfig(
            udf=with_params,
            input=arg_list,
        )


EMPTY_UDF = GeoPandasUdfV2(
    name="EMPTY_UDF", code="", entrypoint="", table_schema=Schema(fields=[])
)

AnyBaseUdf = Annotated[GeoPandasUdfV2, Field(..., discriminator="type")]


class RootAnyBaseUdf(RootModel[AnyBaseUdf]):
    pass


def load_udf_from_response_data(data: dict, context=None) -> RootAnyBaseUdf:
    """Return a UDF from an HTTP response body, adding in metadata if necessary"""

    if context is None:
        context = {}

    udf = RootAnyBaseUdf.model_validate_json(data["udf_body"], context=context).root
    # Restore metadata fields if they were not already present
    if not udf._get_metadata_safe(METADATA_FUSED_ID) and "id" in data:
        udf._set_metadata_safe(METADATA_FUSED_ID, data["id"])
    if not udf._get_metadata_safe(METADATA_FUSED_SLUG) and "slug" in data:
        udf._set_metadata_safe(METADATA_FUSED_SLUG, data["slug"])
    return udf


# Helper class for AST transformation
class RenameFunctionVisitor(ast.NodeTransformer):
    """
    AST visitor to rename a top level function definition AND its usages (ast.Name in calls/loads).
    """

    def __init__(self, old_name: str, new_name: str):
        super().__init__()
        self.old_name = old_name
        self.new_name = new_name
        self.found_definition = False  # Track if the definition was found

    @override
    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        if node.name == self.old_name:
            if self.found_definition:
                warnings.warn(
                    f"Found multiple definitions of entrypoint function '{self.old_name}'"
                )
            node.name = self.new_name
            self.found_definition = True
        # Continue visiting other nodes
        return self.generic_visit(node)

    @override
    def visit_Name(self, node: ast.Name) -> ast.AST:
        # Rename names used in a 'load' context (i.e., references/calls)
        if node.id == self.old_name and isinstance(node.ctx, ast.Load):
            node.id = self.new_name
        return node  # Return the node itself (potentially modified)



from __future__ import annotations

import contextlib
import functools
import hashlib
import inspect
import os
import pickle
import re
import subprocess
import sys
import time
import warnings
from dataclasses import dataclass, fields
from datetime import datetime, timedelta, timezone
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Awaitable, Callable, Iterable, TypeVar
from uuid import uuid4

from loguru import logger

from fused._environment import is_pyodide
from fused._optional_deps import (
    HAS_PANDAS,
    PD_DATAFRAME,
    PD_HASH_OBJECT_FN,
    PD_INDEX,
    PD_SERIES,
)
from fused._options import StorageStr, get_cache_dir
from fused._options import options as OPTIONS
from fused.models.udf import BaseUdf
from fused.warnings import FusedDeprecationWarning

DEFAULT_CACHE_MAX_AGE = "12h"


class CacheResultError(Exception):
    pass


class LockError(Exception):
    pass


@dataclass
class CacheLogEntry:
    expiration: datetime
    uuid: str
    status: str

    # TODO: Make class responsible for serialization to consolidate logic
    @classmethod
    def byte_length(cls):
        ex_length = 26  # datetime.isoformat
        uuid = 36  # str(uuid4())
        status = 1  # char
        commas = len(fields(cls)) - 1
        newline = 1
        return ex_length + uuid + status + commas + newline


@dataclass
class ArgumentParseResult:
    arg_end_index: int = -1
    """Index specifying the end of positional arguments"""
    exclude_var_positional: bool = False
    """Whether variable positional arguments (*param_name) are excluded"""
    param_names: list[str] = None
    """List of all parsed param names defined on the function. Used to specifically map the order of params to their name"""


def print_message(message, verbose=None):
    verbose = verbose if verbose is not None else OPTIONS.verbose_cached_functions
    if verbose:
        print(message, file=sys.stderr)


# TODO: Possibly move all related methods to a class


def _hashify(func) -> str:
    hash_object = hashlib.sha256()
    try:
        if hasattr(func, "__fused_cached_fn"):
            return _hashify(func.__fused_cached_fn)
        elif isinstance(func, BaseUdf):
            # Must be before callable(func) as it is callable but has no __code__ attribute
            hash_object.update(func.model_dump_json().encode("utf-8"))
        elif callable(func):
            hash_object.update(inspect.getsource(func).encode("utf8"))
        else:
            if HAS_PANDAS and isinstance(func, (PD_DATAFRAME, PD_SERIES, PD_INDEX)):
                hash_target = PD_HASH_OBJECT_FN(func).to_numpy()
            else:
                hash_target = str(func).encode("utf-8")

            hash_object.update(hash_target)
        return hash_object.hexdigest()
    except Exception as e:
        logger.warning(f"Error Hashing {e}")
        return ""


def parse_function_arguments(
    func: Callable, excluded_parameters: set
) -> ArgumentParseResult:
    """Parse a cache decorated function's arguments to help with calculating the function's cache hash key.

     Inspect parameters to identify arg names and their index. Also noting the position where normal positional
     arguments end and deciding whether to include variable positional arguments in the hash key.

    Args:
        func: The function to inspect
        excluded_parameters: Set of parameter names to exclude in hash calculation

    Returns:
        ArgumentParseResult
    """
    sig = inspect.signature(func)
    arg_end_index = -1
    exclude_var_positional = False

    param_names = []
    for i, (name, p) in enumerate(sig.parameters.items()):
        if p.kind == p.POSITIONAL_ONLY or p.kind == p.POSITIONAL_OR_KEYWORD:
            arg_end_index = i
        elif p.kind == p.VAR_POSITIONAL and p.name in excluded_parameters:
            exclude_var_positional = True
        param_names.append(name)

    return ArgumentParseResult(arg_end_index, exclude_var_positional, param_names)


def _parse_time_format(t: str | int) -> timedelta:
    if t == 0 or (isinstance(t, str) and t.strip() == "0"):
        return timedelta(0)

    if isinstance(t, int):
        t = f"{t}s"

    quantifier = re.match(r"(\d+)([smhd])$", t.strip().lower())
    if not quantifier:
        raise ValueError(
            f"Invalid time format {t!r}: Use a number followed by one of 's' (seconds), 'm' (minutes), 'h' (hours), or 'd' (days)."
        )
    value, unit = int(quantifier.group(1)), quantifier.group(2)
    if value < 0:
        raise ValueError(f"Time format {t!r} cannot be less than zero")

    delta_map = {
        "s": timedelta(seconds=value),
        "m": timedelta(minutes=value),
        "h": timedelta(hours=value),
        "d": timedelta(days=value),
    }

    return delta_map[unit]


def _format_now_with_delta(delta: timedelta) -> str:
    now = datetime.now(timezone.utc).replace(microsecond=0)
    return (now + delta).isoformat()


def _write_cache_file(
    func: Callable,
    args: tuple,
    kwargs: dict,
    cache_path: Path,
    uuid: str,
    max_age_delta: timedelta,
    timeout_delta: timedelta,
):
    logger.debug(f"Caching {func.__qualname__} under {uuid}")
    in_prog_expires = _format_now_with_delta(timeout_delta)
    # Acquire lock and set status to in-progress
    _write_cache_log(cache_path, uuid, in_prog_expires, status="p")
    try:
        data = func(*args, **kwargs)
    except Exception as e:
        logger.debug(f"Error running {func.__qualname__}: {e}")
        # Release lock and set status to fail
        expires = _format_now_with_delta(timedelta(0))
        _write_cache_log(cache_path, uuid, expires, status="f")
        raise

    with open(cache_path / f"{uuid}.pickle", "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    # Release lock and set status to done
    expires = _format_now_with_delta(max_age_delta)
    _write_cache_log(cache_path, uuid, expires, status="d")
    return data


async def _write_cache_file_async(
    func: Callable,
    args: tuple,
    kwargs: dict,
    cache_path: Path,
    uuid: str,
    max_age_delta: timedelta,
    timeout_delta: timedelta,
):
    logger.debug(f"Caching {func.__qualname__} under {uuid}")
    in_prog_expires = _format_now_with_delta(timeout_delta)
    # Acquire lock and set status to in-progress
    _write_cache_log(cache_path, uuid, in_prog_expires, status="p")
    try:
        data = await func(*args, **kwargs)
    except Exception as e:
        logger.debug(f"Error running {func.__qualname__}: {e}")
        # Release lock and set status to fail
        expires = _format_now_with_delta(timedelta(0))
        _write_cache_log(cache_path, uuid, expires, status="f")
        raise

    with open(cache_path / f"{uuid}.pickle", "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    # Release lock and set status to done
    expires = _format_now_with_delta(max_age_delta)
    _write_cache_log(cache_path, uuid, expires, status="d")
    return data


def _cleanup_expired_links(directory: Path, max_age: timedelta):
    max_age_ns = max_age.total_seconds() * 1e9
    now_ns = time.time_ns()
    for link in directory.iterdir():
        last_modified = int(link.name.split("-", maxsplit=1)[0])
        diff = now_ns - last_modified
        if diff > max_age_ns:
            logger.debug(f"Removing expired {link} link: {diff}, {max_age_ns}")
            link.unlink(missing_ok=True)


def _cleanup_expired_file_locks(directory: Path, max_age: timedelta):
    max_age_ns = max_age.total_seconds() * 1e9
    now_ns = time.time_ns()
    for lock in directory.iterdir():
        with open(lock) as lockfile:
            last_modified = int(lockfile.read())
        diff = now_ns - last_modified
        if diff > max_age_ns:
            logger.debug(f"Removing expired {lock} lock: {diff}, {max_age_ns}")
            lock.unlink(missing_ok=True)


@contextlib.contextmanager
def _get_lock(target: Path, lock_dir, new_cache_uuid):
    if is_pyodide():
        # Pyodide does not support os.link, so we use exclusive file creation
        lock_file = lock_dir / f"{target.name}.lock"
        lock_method = _file_lock(lock_file)
    else:
        now_ns = time.time_ns()
        lock_file = lock_dir / f"{now_ns}-{new_cache_uuid}.log"
        lock_method = _link_based_lock(lock_file, target)

    with lock_method:
        yield


@contextlib.contextmanager
def _file_lock(lock_file: Path):
    # Cleanup expired locks if they were not cleaned up from previous runs
    _cleanup_expired_file_locks(lock_file.parent, timedelta(seconds=5))

    lock_fd = None
    try:
        lock_fd = os.open(lock_file, os.O_CREAT | os.O_EXCL | os.O_RDWR)
        os.write(lock_fd, str(time.time_ns()).encode("utf8"))
        os.close(lock_fd)
        yield
    except FileExistsError as e:
        raise LockError(f"Could not lock {e}")
    finally:
        if lock_fd:
            os.remove(lock_file)


@contextlib.contextmanager
def _link_based_lock(link_file: Path, target: Path):
    logger.debug(f"Attempting to lock {target}")
    # Cleanup expired links if they were not cleaned up from previous runs
    _cleanup_expired_links(link_file.parent, timedelta(seconds=5))

    try:
        # TODO: convert to link_file.hardlink_to(target) once 3.9 is no longer supported
        os.link(target, link_file)
    except OSError as e:
        raise LockError(f"Could not lock {e}")

    # nlink = 2 because we have 1 from the target file, then 1 from the link serving as a lock.
    # nlink > 2 means multiple links were created and target currently "locked".
    if target.stat().st_nlink == 2:
        logger.debug(f"Lock acquired on {target}")
    else:
        link_file.unlink(missing_ok=True)
        raise LockError("Too many locks")

    try:
        yield
    finally:
        logger.debug(f"Unlocking {target}")
        link_file.unlink()


def _write_cache_log(cache_path: Path, uuid: str, expires_on: str, status: str):
    log_file = cache_path / "index.log"

    with open(log_file, "a") as f:
        f.write(f"{expires_on},{uuid},{status}\n")

    logger.debug(f"Wrote {uuid} to cache log {log_file} with status {status}")


def _read_cache_log(cache_path: Path) -> CacheLogEntry | None:
    log_file = cache_path / "index.log"
    if is_pyodide():
        try:
            with open(log_file, "rb") as logfile:
                logfile.seek(CacheLogEntry.byte_length() * -1, 2)
                recent_cache_entry = logfile.read().strip().decode("utf8")
        except OSError as e:
            # file is empty
            if "invalid argument" in str(e).lower():
                recent_cache_entry = None
            else:
                logger.debug(f"Reading log file failed: {e}")
                raise

    else:
        try:
            # TODO: Replace tail with builtin file reads and update log file to use fixed byte length record sizes
            # Strip to ensure we don't break if an empty line exists
            recent_cache_entry = subprocess.check_output(
                ["tail", "-1", str(log_file)], encoding="utf8"
            ).strip()
        except CalledProcessError as e:
            # Something wrong with the file or executing command
            logger.debug(f"Reading log file failed: {e}")
            raise

    if recent_cache_entry:
        try:
            expiration, uuid, status = (
                d.strip() for d in recent_cache_entry.split(",")
            )
            logger.debug(
                f"Recent cache entry found in {log_file}: {expiration!r}, {uuid!r}, {status!r}"
            )
            return CacheLogEntry(datetime.fromisoformat(expiration), uuid, status)
        except (TypeError, ValueError) as e:
            logger.debug(
                f"Issues with parsing last line in log {e}. Truncating logfile {log_file}"
            )
            # Any issue with parsing the last line. This could mean the file is corrupted, therefore we clear the file.
            with open(log_file, "w") as logfile:
                logfile.truncate()

    return None


def _read_cache_file(cache_path: Path, uuid: str):
    with open(cache_path / f"{uuid}.pickle", "rb") as f:
        data = pickle.load(f)
        return data


def wait_for_ready_result(
    fn_name: str,
    start_time: datetime,
    max_retries: int,
    wait_time_seconds: int,
    path_dir: Path,
    *,
    result_expected_soon: bool = False,
    cache_verbose: bool | None = None,
):
    """Wait until a cache entry is ready to be read or immediately return if no result is found.

    A ready result is one that is no longer in-progress of execution or being written by another parallel invocation.
    This can either be successful (cache written and can be read) or failure (function that was decorated with cache
    decorator failed mid-execution). When this is used for reading the cache for the first time, and an expired
    in-progress or no result could exist, we do not want to wait for one to be ready. However, in the case where we
    know another invocation has acquired a lock and will write an entry, `result_expected_soon` can be used during
    these situations to continue to wait.

    Args:
        fn_name (str): Function name to log
        start_time (datetime): Baseline datetime used to check expiration
        max_retries (int): Max amount of attempts to read a valid cache entry
        wait_time_seconds (int): Amount of time to wait per retry
        path_dir (Path): Directory where cache files exist for the function
        result_expected_soon (bool): Whether to wait if an expired in-progress or no cache entry is detected. Useful
            when the caller knows that an entry will soon be available.
        cache_verbose: Print a message when a cached result is returned

    Raises:
        CacheResultError: When max retries has been reached, cache entry has expired, no cache entry exists, or a
            failed entry was detected.
    Returns:
        Un-serialized cache data if a non-expired, successful entry exists
    """

    retries = 0
    now = start_time
    while True:
        log_msg = ""
        log_params = {}
        cache_entry = _read_cache_log(path_dir)
        if cache_entry is None:
            if not result_expected_soon:
                break
            # In the situation where no entry exists and multiple invocations attempt to write to the log file, only one
            # will successfully acquire the lock while the others will wait for the result in this loop. In this case,
            # we want to wait for the result, so we do not break out of the loop.
            log_msg = "Sleeping. No entry found yet. Retry [{retries}/{max_retries}]"
        elif cache_entry.status == "p" and now < cache_entry.expiration:
            log_msg = (
                "Sleeping. Cache entry reporting {cache_status}, {now} < {expiration}, "
                "current write might be in progress. Retry [{retries}/{max_retries}]"
            )
            log_params = {
                "cache_status": cache_entry.status,
                "now": now,
                "expiration": cache_entry.expiration,
            }
        elif cache_entry.status == "p" and now >= cache_entry.expiration:
            # In the situation where a stale in-progress entry exists and multiple invocations attempt to write to the
            # log file, only one will successfully acquire the lock while the others will wait for result in this loop.
            # In this case, we want to wait for the result, so we do not raise a cache error.
            if not result_expected_soon:
                raise CacheResultError(f"Expired: {cache_entry.expiration}")
        else:
            # Cache entry no longer in progress
            break

        retries += 1
        if retries > max_retries:
            raise TimeoutError("Max retries reached")

        if log_msg:
            log_params.update({"retries": retries, "max_retries": max_retries})
            logger.debug(log_msg.format(**log_params))
        time.sleep(wait_time_seconds)

    if cache_entry is None:
        raise CacheResultError("No cache result")

    logger.debug(
        f"Detected status is {cache_entry.status} and expiration is {cache_entry.expiration}"
    )
    if cache_entry.status == "d" and now < cache_entry.expiration:
        logger.debug(f"Reading cache for {fn_name} under {cache_entry.uuid}")
        print_message(f"{fn_name}() returned cached result", verbose=cache_verbose)
        # Cache exists and valid
        return _read_cache_file(path_dir, cache_entry.uuid)
    raise CacheResultError(
        f"Failed status or expired: {cache_entry.uuid}. {cache_entry.status}, {cache_entry.expiration}"
    )


def _cache(
    func: Callable,
    *args,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    cache_reset: bool | None = None,
    storage: StorageStr = "auto",
    cache_key_exclude: Iterable[str] = None,
    cache_verbose: bool | None = None,
    **kwargs,
) -> Any:
    """Internal method used by cache decorator to cache a function's pickleable response to a file.

    Args:
        func: The decorated function.
        *args: Positional arguments to the decorated function.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m),
            hours (h), and days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time for concurrent calls to wait for the decorated function
            to finish execution and to write the cache file currently being written by another concurrent call.
            Waiting will end before the timeout if a finished cache file is detected after reading the most recent entry
            in the log. Otherwise, after the timeout, it will either find a finished cached file to read, or it will
            write a new one if a cache file has not been recorded in the log.
        cache_reset: Ignore `cache_max_age` and overwrite cached result.
        storage: Set where the cache data is stored. Supported values are "auto", "mount" and "local". Auto will
            automatically select the storage location defined in options (mount if it exists, otherwise local) and
            ensures that it exists and is writable. Mount gets shared across executions where local will only be shared
            within the same execution.
        cache_key_exclude: An iterable of parameter names to exclude from the cache key calculation. Useful for
            arguments that do not affect the result of the function and could cause unintended cache expiry (e.g.
            database connection objects)
        cache_verbose: Print a message when a cached result is returned
        **kwargs: Keyword arguments to the decorated function.

    Returns:
        Any: Result from the cached function
    """
    # Calculate expires_on
    max_age_delta = _parse_time_format(cache_max_age)
    timeout_delta = _parse_time_format(concurrent_lock_timeout)

    try:
        path = _get_cache_base_path(path, storage)
        # TODO: ignore `_`

        if cache_key_exclude is None:
            cache_key_exclude = set()
        elif isinstance(cache_key_exclude, str):
            cache_key_exclude = {cache_key_exclude}
        elif isinstance(cache_key_exclude, bytes):
            raise ValueError(
                "'cache_key_exclude' must be a collection of parameter names or a single name as a string."
            )
        else:
            cache_key_exclude = set(cache_key_exclude)

        # 1. Hashify function
        id = _hashify(func)

        parsed_result = parse_function_arguments(func, cache_key_exclude)

        # 2. Hashify args
        for i, v in enumerate(args):
            # Any args after arg_end_index are variable positional
            if i > parsed_result.arg_end_index and parsed_result.exclude_var_positional:
                break

            if parsed_result.param_names[i] not in cache_key_exclude:
                id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            if k not in cache_key_exclude:
                id += k + _hashify(kwargs[k])

        # 4. Hashify cache_max_age
        id += _hashify(max_age_delta)

        # 5. Hashify composite id
        id = _hashify(id)

        path_dir = path / f"data_{id}"
        locks_dir = path / f"data_{id}" / "locks"

        path_dir.mkdir(parents=True, exist_ok=True)
        locks_dir.mkdir(parents=True, exist_ok=True)
        log_file = path_dir / "index.log"
        # Ensure log file exists
        log_file.touch()

        new_cache_uuid = str(uuid4())
        timeout_seconds = timeout_delta.total_seconds()
        now = datetime.now(timezone.utc)
        wait_time_seconds = 1
        max_retries = int(timeout_seconds // wait_time_seconds)
        fn_name = func.__name__

        if not cache_reset:
            try:
                return wait_for_ready_result(
                    fn_name,
                    now,
                    max_retries,
                    wait_time_seconds,
                    path_dir,
                    cache_verbose=cache_verbose,
                )
            except (CacheResultError, TimeoutError) as e:
                logger.debug(f"Cache entry not valid: {e}")
                # A completed cache result could not be found and needs to be written. Continue to write logic.
                pass

        # Multiple simultaneous write calls can happen. We attempt to lock and those that fail will wait for the cache
        # entry to complete. This is extremely helpful for cached functions that need a "run once" and perform
        # operations in a non-parallel safe way.
        try:
            with _get_lock(path_dir / "index.log", locks_dir, new_cache_uuid):
                return _write_cache_file(
                    func,
                    args,
                    kwargs,
                    path_dir,
                    new_cache_uuid,
                    max_age_delta,
                    timeout_delta,
                )
        except LockError as e:
            logger.debug(f"LockError: {e}")

            # Lock currently in use to write cache entry, wait for result.
            # If invocation that has the lock errors due to an error in the decorated function or otherwise, this will
            # either time out or an exception will be raised that a failed result was detected.
            return wait_for_ready_result(
                fn_name,
                now,
                max_retries,
                wait_time_seconds,
                path_dir,
                result_expected_soon=True,
                cache_verbose=cache_verbose,
            )

    except Exception as e:
        logger.debug(f"Error Caching {e}")
        if isinstance(e, TimeoutError):
            msg = f"Wait time exceeded concurrent_lock_timeout of {timeout_delta.total_seconds()}s"
            raise TimeoutError(msg) from e
        raise e


async def _cache_async(
    func: Callable,
    *args,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    cache_reset: bool | None = None,
    storage: StorageStr = "auto",
    cache_key_exclude: Iterable[str] = None,
    cache_verbose: bool | None = None,
    **kwargs,
) -> Any:
    """Async internal method used by cache decorator to cache a function's pickleable response to a file.

    Args:
        func: The decorated function.
        *args: Positional arguments to the decorated function.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m),
            hours (h), and days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time for concurrent calls to wait for the decorated function
            to finish execution and to write the cache file currently being written by another concurrent call.
            Waiting will end before the timeout if a finished cache file is detected after reading the most recent entry
            in the log. Otherwise, after the timeout, it will either find a finished cached file to read, or it will
            write a new one if a cache file has not been recorded in the log.
        cache_reset: Ignore `cache_max_age` and overwrite cached result.
        storage: Set where the cache data is stored. Supported values are "auto", "mount" and "local". Auto will
            automatically select the storage location defined in options (mount if it exists, otherwise local) and
            ensures that it exists and is writable. Mount gets shared across executions where local will only be shared
            within the same execution.
        cache_key_exclude: An iterable of parameter names to exclude from the cache key calculation. Useful for
            arguments that do not affect the result of the function and could cause unintended cache expiry (e.g.
            database connection objects)
        cache_verbose: Print a message when a cached result is returned
        **kwargs: Keyword arguments to the decorated function.

    Returns:
        Any: Result from the cached function
    """
    # Calculate expires_on
    max_age_delta = _parse_time_format(cache_max_age)
    timeout_delta = _parse_time_format(concurrent_lock_timeout)

    try:
        path = _get_cache_base_path(path, storage)
        # TODO: ignore `_`

        if cache_key_exclude is None:
            cache_key_exclude = set()
        elif isinstance(cache_key_exclude, str):
            cache_key_exclude = {cache_key_exclude}
        elif isinstance(cache_key_exclude, bytes):
            raise ValueError(
                "'cache_key_exclude' must be a collection of parameter names or a single name as a string."
            )
        else:
            cache_key_exclude = set(cache_key_exclude)

        # 1. Hashify function
        id = _hashify(func)

        parsed_result = parse_function_arguments(func, cache_key_exclude)

        # 2. Hashify args
        for i, v in enumerate(args):
            # Any args after arg_end_index are variable positional
            if i > parsed_result.arg_end_index and parsed_result.exclude_var_positional:
                break

            if parsed_result.param_names[i] not in cache_key_exclude:
                id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            if k not in cache_key_exclude:
                id += k + _hashify(kwargs[k])

        # 4. Hashify cache_max_age
        id += _hashify(max_age_delta)

        # 5. Hashify composite id
        id = _hashify(id)

        path_dir = path / f"data_{id}"
        locks_dir = path / f"data_{id}" / "locks"

        path_dir.mkdir(parents=True, exist_ok=True)
        locks_dir.mkdir(parents=True, exist_ok=True)
        log_file = path_dir / "index.log"
        # Ensure log file exists
        log_file.touch()

        new_cache_uuid = str(uuid4())
        timeout_seconds = timeout_delta.total_seconds()
        now = datetime.now(timezone.utc)
        wait_time_seconds = 1
        max_retries = int(timeout_seconds // wait_time_seconds)
        fn_name = func.__qualname__
        if not cache_reset:
            try:
                return wait_for_ready_result(
                    fn_name,
                    now,
                    max_retries,
                    wait_time_seconds,
                    path_dir,
                    cache_verbose=cache_verbose,
                )
            except (CacheResultError, TimeoutError) as e:
                logger.debug(f"Cache entry not valid: {e}")
                # A completed cache result could not be found and needs to be written. Continue to write logic.
                pass

        # Multiple simultaneous write calls can happen. We attempt to lock and those that fail will wait for the cache
        # entry to complete. This is extremely helpful for cached functions that need a "run once" and perform
        # operations in a non-parallel safe way.
        try:
            with _get_lock(path_dir / "index.log", locks_dir, new_cache_uuid):
                return await _write_cache_file_async(
                    func,
                    args,
                    kwargs,
                    path_dir,
                    new_cache_uuid,
                    max_age_delta,
                    timeout_delta,
                )
        except LockError as e:
            logger.debug(f"LockError: {e}")

            # Lock currently in use to write cache entry, wait for result.
            # If invocation that has the lock errors due to an error in the decorated function or otherwise, this will
            # either time out or an exception will be raised that a failed result was detected.
            return wait_for_ready_result(
                fn_name,
                now,
                max_retries,
                wait_time_seconds,
                path_dir,
                result_expected_soon=True,
                cache_verbose=cache_verbose,
            )

    except Exception as e:
        logger.debug(f"Error Caching {e}")
        if isinstance(e, TimeoutError):
            msg = f"Wait time exceeded concurrent_lock_timeout of {timeout_delta.total_seconds()}s"
            raise TimeoutError(msg) from e
        raise e


def _cache_internal(func, **decorator_kwargs):
    # Handle kwargs passed from decorator
    if decorator_kwargs.get("reset") is not None:
        warnings.warn(
            "The 'reset' keyword is deprecated. Use 'cache_reset' instead",
            FusedDeprecationWarning,
        )
        decorator_kwargs["cache_reset"] = decorator_kwargs["reset"]
    decorator_kwargs.pop("reset", None)

    def decorator(func):
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper_async(*args, **kwargs):
                # Handle kwargs passed from function
                cache_reset_not_passed = (
                    "cache_reset" not in kwargs
                    and decorator_kwargs["cache_reset"] is None
                )
                if cache_reset_not_passed and kwargs.get("reset") is not None:
                    warnings.warn(
                        "The 'reset' keyword is deprecated. Use 'cache_reset' instead",
                        FusedDeprecationWarning,
                    )
                    kwargs["cache_reset"] = kwargs["reset"]
                    kwargs.pop("reset", None)

                # Allow passing cache kwargs through fn kwargs. decorator_kwargs are specified from cache decorator
                for k in decorator_kwargs:
                    if k in kwargs:
                        decorator_kwargs[k] = kwargs.pop(k)

                return await _cache_async(
                    func,
                    *args,
                    **decorator_kwargs,
                    **kwargs,
                )

            wrapper_async.__fused_cached_fn = func

            return wrapper_async
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # Handle kwargs passed from function
                cache_reset_not_passed = (
                    "cache_reset" not in kwargs
                    and decorator_kwargs["cache_reset"] is None
                )
                if cache_reset_not_passed and kwargs.get("reset") is not None:
                    warnings.warn(
                        "The 'reset' keyword is deprecated. Use 'cache_reset' instead",
                        FusedDeprecationWarning,
                    )
                    kwargs["cache_reset"] = kwargs["reset"]
                    kwargs.pop("reset", None)
                # Allow passing cache kwargs through fn kwargs. decorator_kwargs are specified from cache decorator
                for k in decorator_kwargs:
                    if k in kwargs:
                        decorator_kwargs[k] = kwargs.pop(k)

                return _cache(
                    func,
                    *args,
                    **decorator_kwargs,
                    **kwargs,
                )

            wrapper.__fused_cached_fn = func

            return wrapper

    if callable(func):  # w/o args
        return decorator(func)
    else:  # w/ args
        return decorator


def cache(
    func: Callable[..., Any] | None = None,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    cache_reset: bool | None = None,
    storage: StorageStr = "auto",
    cache_key_exclude: Iterable[str] = None,
    cache_verbose: bool | None = None,
    reset: bool | None = None,
) -> Callable[..., Any]:
    """Decorator to cache the return value of a function.

    This function serves as a decorator that can be applied to any function
    to cache its return values. The cache behavior can be customized through
    keyword arguments.

    Args:
        func (Callable, optional): The function to be decorated. If None, this
            returns a partial decorator with the passed keyword arguments.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m), hours (h), and
            days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time in seconds for subsequent concurrent calls to wait for a previous
            concurrent call to finish execution and to write the cache file.
        cache_reset: Ignore `cache_max_age` and overwrite cached result.
        storage: Set where the cache data is stored. Supported values are "auto", "mount" and "local". Auto will
            automatically select the storage location defined in options (mount if it exists, otherwise local) and
            ensures that it exists and is writable. Mount gets shared across executions where local will only be shared
            within the same execution.
        cache_key_exclude: An iterable of parameter names to exclude from the cache key calculation. Useful for
            arguments that do not affect the result of the function and could cause unintended cache expiry (e.g.
            database connection objects)
        cache_verbose: Print a message when a cached result is returned
    Returns:
        Callable: A decorator that, when applied to a function, caches its
        return values according to the specified keyword arguments.

    Examples:
        Use the `@cache` decorator to cache the return value of a function in a custom path.

        ```py
        @cache(path="/tmp/custom_path/")
        def expensive_function():
            # Function implementation goes here
            return result
        ```

        If the output of a cached function changes, for example if remote data is modified,
        it can be reset by running the function with the `cache_reset` keyword argument. Afterward,
        the argument can be cleared.

        ```py
        @cache(path="/tmp/custom_path/", cache_reset=True)
        def expensive_function():
            # Function implementation goes here
            return result
        ```
    """
    return _cache_internal(
        func=func,
        cache_max_age=cache_max_age,
        path=path,
        concurrent_lock_timeout=concurrent_lock_timeout,
        cache_reset=cache_reset,
        storage=storage,
        cache_key_exclude=cache_key_exclude,
        cache_verbose=cache_verbose,
        reset=reset,
    )


T = TypeVar("T")


def cache_call(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Directly calls a function with caching.

    This function directly calls the provided function with the given arguments
    and keyword arguments, caching its return value. The cache used depends on
    the implementation of the `_cache` function.

    Args:
        func (Callable): The function to call and cache its result.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The cached return value of the function.

    Raises:
        Exception: Propagates any exception raised by the function being called
        or the caching mechanism.
    """
    return _cache(func, *args, **kwargs)


async def cache_call_async(
    func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> T:
    """Asynchronously calls a function with caching.

    Similar to `cache_call`, but for asynchronous functions. This function
    awaits the provided async function, caches its return value, and then
    returns it. The specifics of the caching mechanism depend on the
    implementation of `_cache_async`.

    Args:
        func (Callable): The asynchronous function to call and cache its result.
        *args: Positional arguments to pass to the async function.
        **kwargs: Keyword arguments to pass to the async function.

    Returns:
        The cached return value of the async function.

    Raises:
        Exception: Propagates any exception raised by the async function being
        called or the caching mechanism.

    Examples:
        async def fetch_data(param):
            # Async function implementation goes here
            return data

        # Usage
        result = await cache_call_async(fetch_data, 'example_param')
    """
    return await _cache_async(func, *args, **kwargs)


def _get_cache_base_path(sub_path, storage: StorageStr = "auto") -> Path:
    sub_path = sub_path.strip("/")
    # TODO: consider udf name in path once available from Fused global context
    base_path = OPTIONS.cache_directory
    if storage != "auto":
        base_path = get_cache_dir(storage)

    sub_path = base_path / sub_path

    return sub_path
